// SortListBox.cpp : implementation file
// Copyright (c) 2016 by David Wincelberg
//
// Window styles:
// Recommended: horizontal scroll, vertical scroll, disable no scroll
// Optional: sort
//
// For an owner-draw list box with strings, select has-strings and
// implement DrawItem and MeasureItem.  CompareItem will not be called.
//
// Includes horizontal scrolling (based on the following reference).
//
// Reference:
// "How to implement horizontal scrolling in a CListBox-derived class" at:
// https://support.microsoft.com/en-us/kb/146437
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SortListBox.h"

#include "ListBoxData.h"
#include "ListBoxOrder.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSortListBox

// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortListBox::CSortListBox (CListBoxOrder* pListBoxOrder /* = nullptr */,
							bool bHandleDelete /* = true */)
{
	m_pListBoxOrder  = (pListBoxOrder == nullptr ? new CListBoxOrder() : pListBoxOrder);
	m_bHandleDelete  = bHandleDelete;
	m_bSorting       = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;

	// Stores pointer for horizontal-extent calculations.
	SetBoxPointer (this);
}	// CSortListBox


// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortListBox::CSortListBox (CStringOrder* pStringOrder,
							bool bHandleDeleteInner /* = true */,
							bool bHandleDeleteOuter /* = true */)
{
	m_pListBoxOrder  = new CListBoxOrder (pStringOrder, bHandleDeleteInner);
	m_bHandleDelete  = bHandleDeleteOuter;
	m_bSorting       = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;

	// Stores pointer for horizontal-extent calculations.
	SetBoxPointer (this);
}	// CSortListBox


CSortListBox::~CSortListBox()
{
	if (m_bHandleDelete)
		delete m_pListBoxOrder;
}	// ~CSortListBox


BEGIN_MESSAGE_MAP(CSortListBox, CListBox)
	//{{AFX_MSG_MAP(CSortListBox)
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
	ON_MESSAGE (LB_SETTABSTOPS, OnSetTabStops)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSortListBox message handlers

// Action:	Checks the styles and sets the sorting flag.
void CSortListBox::PreSubclassWindow()
{
	DWORD dwStyle = GetStyle();
	if ((dwStyle & (LBS_OWNERDRAWFIXED | LBS_OWNERDRAWVARIABLE)) != 0
	&& (dwStyle & LBS_HASSTRINGS) == 0) {
		TRACE (_T("** For CSortListBox, set style has-strings with an owner-draw style.\n"));
		ASSERT (false);
	}

	m_bSorting = ((dwStyle & LBS_SORT) != 0);

	CListBox::PreSubclassWindow();
}	// PreSubclassWindow


// Action:	Replaces the sort-order object.
//			Deletes the old one unless m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortListBox::SetSortOrder (CListBoxOrder* pListBoxOrder,
								 bool bHandleDelete /* = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the sort-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pListBoxOrder == nullptr)
		return;

	if (m_bHandleDelete)
		delete m_pListBoxOrder;

	m_pListBoxOrder = pListBoxOrder;
	m_bHandleDelete = bHandleDelete;
}	// SetSortOrder


// Action:	Replaces the string-order object.
//			Deletes the old one unless CListBoxOrder's m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortListBox::SetStringOrder (CStringOrder* pStringOrder,
								   bool bHandleDelete /*  = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the string-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pStringOrder == nullptr || m_pListBoxOrder == nullptr)
		return;

	m_pListBoxOrder->SetStringOrder (pStringOrder, bHandleDelete);
}	// SetStringOrder


// Action:	Checks if a list-box-data object was supplied; o.w. allocates one.
bool CSortListBox::CheckListBoxData (CListBoxData** ppListBoxData)
{
	if (*ppListBoxData == nullptr) {
		try {
			*ppListBoxData = new CListBoxData();
		}
		catch (CMemoryException* e) {
			e->Delete();
			return false;
		}
	}

	return true;
}	// CheckListBoxData


// Action:	Adds a string to the list box and stores extent data.
//			If sorting is selected, sorts based on the current string order.
// Note:	Allows for storing a data item that is in a subclass of CListBoxData.
int CSortListBox::AddString (LPCTSTR lpszItem, CListBoxData* pListBoxData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListBoxData;
		return LB_ERR;
	}

	if (!CheckListBoxData (&pListBoxData))
		return LB_ERR;

	if (!m_bSorting) {
		int nResult = CListBox::AddString (lpszItem);
		if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
			delete pListBoxData;
			return nResult;
		}

		StoreExtent (nResult, lpszItem, pListBoxData);
		m_bAllowOrderMod = false;
		return nResult;
	}

	int nPosition = 0, nCount = GetCount();

	if (nCount > 0 && m_pListBoxOrder != nullptr) {
		try {
			CString strItem = lpszItem;
			nPosition = SortStrings (strItem, pListBoxData, nCount);
		}
		catch (CMemoryException* e) {
			e->Delete();
			return LB_ERR;
		}
	}

	int nResult = CListBox::InsertString (nPosition, lpszItem);
	if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
		delete pListBoxData;
		return nResult;
	}

	StoreExtent (nResult, lpszItem, pListBoxData);
	m_bAllowOrderMod = false;
	return nResult;
}	// AddString


// Action:	Adds a string to a sorted list or one with insertions.
// Throws:	CMemoryException: =, GetText
int CSortListBox::SortStrings (CString& strItem1, CListBoxData* pListBoxData1, int nCount)
{
	CString strItem2;
	CListBoxData* pListBoxData2;

	if (m_bSorted) {
		// Binary search insertion
		int nLower = 0, nUpper = nCount - 1, nMid, nCompare;

		while (nLower <= nUpper) {
			nMid = (nLower + nUpper) / 2;
			GetText (nMid, strItem2);
			pListBoxData2 = GetItemDataObjPtr (nMid);

			nCompare = m_pListBoxOrder->Compare (strItem1, pListBoxData1,
				strItem2, pListBoxData2);

			if (nCompare > 0)
				nLower = nMid + 1;
			else if (nCompare < 0)
				nUpper = nMid - 1;
			else
				break;
		}

		// When a match is not found:
		// nLower <= position <= nUpper + 1 == nLower
		// See http://www.FileJockeySoftware.com/ANO/BinSearchInsert.html.
		return (nCompare == 0 ? nMid : nLower);
	}
	else {
		// Linear search insertion: only considers added strings
		for (int n = 0;  n < nCount;  ++n) {
			pListBoxData2 = GetItemDataObjPtr (n);

			if (pListBoxData2 == nullptr)
				return -1;
			else if (pListBoxData2->WasAdded()) {
				GetText (n, strItem2);

				if (m_pListBoxOrder->Compare (strItem1, pListBoxData1,
				strItem2, pListBoxData2) < 0)
					return n;
			}
		}

		return -1;	// after last string
	}
}	// SortStrings


// Action:	Inserts a string into a list box at a specified position.
// Note:	Allows for storing a data item that is in a subclass of CListBoxData.
int CSortListBox::InsertString (int nIndex, LPCTSTR lpszItem,
								CListBoxData* pListBoxData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListBoxData;
		return LB_ERR;
	}

	if (!CheckListBoxData (&pListBoxData))
		return LB_ERR;

	int nResult = CListBox::InsertString (nIndex, lpszItem);
	if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
		delete pListBoxData;
		return nResult;
	}

	pListBoxData->SetInserted();
	StoreExtent (nResult, lpszItem, pListBoxData);
	m_bSorted        = false;
	m_bAllowOrderMod = false;
	return nResult;
}	// InsertString


// Action:	Deletes a string, may adjust the horizontal extent and
//			may allow the sort and string orders to change.
int CSortListBox::DeleteString (UINT nIndex)
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);
	int nResult;

	if (pListBoxData != nullptr) {
		int nExtent = pListBoxData->GetExtent();
		nResult = CListBox::DeleteString (nIndex);
		if (nResult == LB_ERR)
			return LB_ERR;

		bool bRecalc = (nExtent >= m_nMaxExtent);

		if (m_bDefer)
			m_bNeedRecalc |= bRecalc;
		else if (bRecalc)
			UpdateMaxExtent();
	}
	else
		nResult = CListBox::DeleteString (nIndex);

	if (GetCount() == 0)
		ResetContent();

	return nResult;
}	// DeleteString


// Action:	Releases memory that is no longer needed.
// Note:	The subclass destructor will be called.
void CSortListBox::DeleteItem (LPDELETEITEMSTRUCT lpDeleteItemStruct)
{
	delete reinterpret_cast<CListBoxData*> (lpDeleteItemStruct->itemData);
}	// DeleteItem


// Action:	Releases memory for all items.
// Note:	DeleteItem is not called since item data are set to zero.
void  CSortListBox::DeleteAllItems()
{
	int nCount = GetCount();
	DWORD dwData;

	for (int n = 0;  n < nCount;  ++n) {
		dwData = CListBox::GetItemData (n);

		if (dwData != LB_ERR) {
			delete reinterpret_cast<CListBoxData*> (dwData);
			CListBox::SetItemData (n, 0);
		}
	}
}	// DeleteAllItems


// Action:	Releases memory, clears the list box, resets the horizontal extent
//			and allows for the sort and string orders to be changed.
void CSortListBox::ResetContent()
{
	DeleteAllItems();
	CListBox::ResetContent();
	SetHorizontalExtent (0);
	m_nMaxExtent     = 0;
	m_bDefer         = false;
	m_bNeedRecalc    = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;
}	// ResetContent


// Action:	Releases memory when this window is destroyed.
void CSortListBox::OnDestroy()
{
	DeleteAllItems();
	CListBox::OnDestroy();
}	// OnDestroy


// Action:	Stores user data in a CListBoxData object.
int CSortListBox::SetItemData (int nIndex, DWORD dwItemData)
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr) {
		pListBoxData->SetItemData (dwItemData);
		return TRUE;
	}
	else
		return LB_ERR;
}	// SetItemData


// Action:	Retrieves user data from a CListBoxData object.
DWORD CSortListBox::GetItemData (int nIndex) const
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr)
		return pListBoxData->GetItemData();
	else
		return LB_ERR;
}	// GetItemData


// Action:	Stores user data in a CListBoxData object.
int CSortListBox::SetItemDataPtr (int nIndex, void* pData)
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr) {
		pListBoxData->SetItemData (reinterpret_cast<DWORD>(pData));
		return TRUE;
	}
	else
		return LB_ERR;
}	// SetItemDataPtr


// Action:	Retrieves user data from a CListBoxData object.
void* CSortListBox::GetItemDataPtr (int nIndex) const
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr)
		return reinterpret_cast<void*> (pListBoxData->GetItemData());
	else
		return reinterpret_cast<void*> (LB_ERR);
}	// GetItemDataPtr


// Action:	Stores user data as a pointer to a CListBoxData object.
// Note:	Copies the data values from the old data object and deletes it.
//			pListBoxData should be from a subclass of CListBoxData.
int CSortListBox::SetItemDataObjPtr (int nIndex, CListBoxData* pListBoxData)
{
	if (pListBoxData == nullptr)
		return LB_ERR;

	CListBoxData* pOldListBoxData = GetItemDataObjPtr (nIndex);

	if (pOldListBoxData != nullptr) {
		pListBoxData = pOldListBoxData;
		delete pOldListBoxData;
		return CListBox::SetItemData (nIndex, reinterpret_cast<DWORD> (pListBoxData));
	}
	else
		return LB_ERR;
}	// SetItemDataObjPtr


// Action:	Retrieves user data as a pointer to a CListBoxData object.
CListBoxData* CSortListBox::GetItemDataObjPtr (int nIndex) const
{
	int nCount = GetCount();

	if (nCount == LB_ERR || nIndex < 0 || nIndex >= nCount)
		return nullptr;

	DWORD dwData = CListBox::GetItemData (nIndex);

	if (dwData != LB_ERR)
		return reinterpret_cast<CListBoxData*> (dwData);
	else
		return nullptr;
}	// GetItemDataObjPtr

/////////////////////////////////////////////////////////////////////////////
// Extent function

// Action:	Intercepts the LB_SETTABSTOPS message to update the tab stop array.
LRESULT CSortListBox::OnSetTabStops (WPARAM wParam, LPARAM lParam)
{
	LRESULT lResult = Default();
	if (!lResult)
		return lResult;

	return TabStopsSet (wParam, lParam, lResult);
}	// OnSetTabStops
